/*
	category.c



	2003  Plus Huang
 */

#include <stdlib.h>
#include "category.h"
#include "category_store.h"
#include "download_thread.h"

Category* category_new ()
{
	Category* cate = g_malloc (sizeof (Category));

	// initial base class category_setting
	category_setting_init (&cate->setting);

	g_static_rec_mutex_init (&cate->mutex);

	download_store_init (&cate->waiting_store);
	download_store_init (&cate->completed_store);
	download_store_init (&cate->recycled_store);

	download_store_set_historical (&cate->completed_store, TRUE);
	download_store_set_historical (&cate->recycled_store, TRUE);

	download_list_view_init (&cate->waiting_view, &cate->waiting_store);
	download_list_view_init (&cate->completed_view, &cate->completed_store);
	download_list_view_init (&cate->recycled_view, &cate->recycled_store);
	download_list_view_hide_runtime_state (&cate->completed_view);
	download_list_view_hide_runtime_state (&cate->recycled_view);

	cate->valid = FALSE;
	cate->reference_count = 1;
	cate->thread_list     = NULL;
	cate->n_thread        = 0;

	category_apply_setting (cate);

	return cate;
}

Category* category_new_from_setting (CategorySetting* cd)
{
	Category* cate = category_new ();

	category_setting_assign (&cate->setting, cd);
	category_apply_setting (cate);

	return cate;
}

void category_ref (Category* ca)
{
	category_lock (ca);
	ca->reference_count++;
	category_unlock (ca);
}

void category_unref (Category* ca)
{
	int  count;

	category_lock (ca);
	count = --ca->reference_count;
	category_unlock (ca);

	if (count > 0)
		return;

	g_static_rec_mutex_free (&ca->mutex);

	category_setting_clear (&ca->setting);

	download_list_view_clear (&ca->waiting_view);
	download_list_view_clear (&ca->completed_view);
	download_list_view_clear (&ca->recycled_view);

	download_store_clear (&ca->waiting_store);
	download_store_clear (&ca->completed_store);
	download_store_clear (&ca->recycled_store);

	g_free (ca);
}

void category_apply_setting (Category* ca)
{
	CategorySetting*  setting = CATEGORY_SETTING (ca);
	DownloadListView* dlv = &ca->waiting_view;

	gtk_tree_view_column_set_visible (dlv->col_completed,
	                                  setting->col_completed);
	gtk_tree_view_column_set_visible (dlv->col_total,
	                                  setting->col_total);
	gtk_tree_view_column_set_visible (dlv->col_speed,
	                                  setting->col_speed);
	gtk_tree_view_column_set_visible (dlv->col_retry,
	                                  setting->col_retry);
	gtk_tree_view_column_set_visible (dlv->col_url,
	                                  setting->col_url);

	gtk_tree_view_column_set_visible (ca->completed_view.col_url,
	                                  setting->col_url);
	gtk_tree_view_column_set_visible (ca->recycled_view.col_url,
	                                  setting->col_url);

	ca->completed_store.capacity_limit = setting->completed_capacity;
	ca->recycled_store.capacity_limit = setting->recycled_capacity;
}

void category_changed_with_children (Category* ca, gint flag_child)
{
	GtkTreePath* path;
	GtkTreeIter  iter;
	CategoryStore* category_store = (CategoryStore*)ca->category_store;

	if (ca->valid==FALSE)
		return;

	// list store
	path = gtk_tree_model_get_path ((GtkTreeModel*)category_store->list_store,
	                                &ca->iter_list_store);
	gtk_tree_model_row_changed ((GtkTreeModel*)category_store->list_store,
	                            path, &ca->iter_list_store);
	gtk_tree_path_free (path);

	// tree store
	path = gtk_tree_model_get_path ((GtkTreeModel*)category_store->tree_store,
	                                &ca->iter_tree_store);
	gtk_tree_model_row_changed ((GtkTreeModel*)category_store->tree_store,
	                            path, &ca->iter_tree_store);
	gtk_tree_path_free (path);

	if (flag_child) {
		gtk_tree_model_iter_children ((GtkTreeModel*)category_store->tree_store,
		                              &iter, &ca->iter_tree_store);
		if (flag_child & CATEGORY_COMPLETED) {
			path = gtk_tree_model_get_path ((GtkTreeModel*)category_store->tree_store,
			                                &iter);
			gtk_tree_model_row_changed ((GtkTreeModel*)category_store->tree_store,
			                            path, &iter);
			gtk_tree_path_free (path);
		}
		if (flag_child & CATEGORY_RECYCLED) {
			gtk_tree_model_iter_next ((GtkTreeModel*)category_store->tree_store,
			                          &iter);
			path = gtk_tree_model_get_path ((GtkTreeModel*)category_store->tree_store,
			                                &iter);
			gtk_tree_model_row_changed ((GtkTreeModel*)category_store->tree_store,
			                            path, &iter);
			gtk_tree_path_free (path);
		}
	}
}

gint category_thread_count_request (Category* ca)
{
	return CATEGORY_SETTING(ca)->n_active_download - ca->n_thread;
}

void category_thread_destroy (Category* ca)
{
	DownloadThread* dt;

	category_lock (ca);

	for (dt=ca->thread_list; dt; dt=dt->next)
		download_thread_destroy (dt);

	category_unlock (ca);
}

gboolean category_thread_activate (Category* ca)
{
	GtkTreeIter     iter;
	DownloadThread* thread;
	gboolean        found;
	gboolean        activated = FALSE;

	do {
		if (category_thread_count_request (ca) <= 0 )
			break;

		download_store_lock (&ca->waiting_store);

		if (found=download_store_find_waiting (&ca->waiting_store, &iter)) {
			thread = download_thread_new (ca, &iter);
			download_store_unlock (&ca->waiting_store);
			download_thread_activate (thread);
		}
		else
			download_store_unlock (&ca->waiting_store);
		activated = TRUE;
	} while (found);

	return activated;
}

void category_thread_stop (Category* ca)
{
	DownloadThread* dt;

	category_lock (ca);

	for (dt=ca->thread_list; dt; dt=dt->next)
		download_thread_stop (dt);

	category_unlock (ca);
}

void category_thread_remove (Category* category,
                             gpointer  th)
{
	DownloadThread* thread = th;
	gboolean        activated = TRUE;

	category_lock (category);

	if (category->thread_list == thread) {
		category->thread_list = thread->next;
	}
	if (thread->next)
		thread->next->prev = thread->prev;
	if (thread->prev)
		thread->prev->next = thread->next;

	category->n_thread--;
	if (category->n_thread==0) {
		CATEGORY_SETTING (category)->state = CATEGORY_STATE_WAITING;
		category_changed (category);
		activated = FALSE;
	}

	category_unlock (category);

	// avoid dead lock
	if (activated == FALSE) {
		category_store_thread_unref ((CategoryStore*)category->category_store);
		category_store_thread_activate_next ((CategoryStore*)category->category_store,
		                                     category);
	}
}

void category_thread_add (Category* category,
                          gpointer  thread)
{
	gboolean activated = FALSE;

	category_lock (category);

	if (category->n_thread==0) {
		CATEGORY_SETTING (category)->state = CATEGORY_STATE_EXECUTING;
		category_changed (category);
		activated = TRUE;
	}
	category->n_thread++;

	if (category->thread_list) {
		((DownloadThread*)category->thread_list)->prev = thread;
		((DownloadThread*)thread)->next = category->thread_list;
	}
	category->thread_list = thread;

	category_unlock (category);

	// avoid dead lock
	if (activated)
		category_store_thread_ref ((CategoryStore*)category->category_store);
}

