#include <stdio.h>
#include "conf_parser.h"

static ConfParserEntry dummy_parser = {
	NULL, NULL, NULL, NULL, NULL
};

void conf_parser_start_element (GMarkupParseContext* context,
                                const gchar*         element_name,
                                const gchar**        attr_names,
                                const gchar**        attr_values,
                                ConfParser*          cp,
                                GError**             error)
{
	conf_parser_start_element_func   start_element = cp->parser->start_element;

	if (start_element)
		start_element (cp, element_name, attr_names, attr_values, cp->data, error);
}

void conf_parser_end_element (GMarkupParseContext* context,
                              const gchar*         element_name,
                              ConfParser*          cp,
                              GError**             error)
{
	conf_parser_end_element_func   end_element = cp->parser->end_element;

	if (end_element)
		end_element (cp, element_name, cp->data, error);
}

void conf_parser_text (GMarkupParseContext* context,
                       const gchar*         text,
                       gsize                text_len,
                       ConfParser*          cp,
                       GError**             error)
{
	conf_parser_text_func   parse_text = cp->parser->text;

	if (parse_text)
		parse_text (cp, text, text_len, cp->data, error);
}

void conf_parser_passthrough (GMarkupParseContext* context,
                              const gchar*         passthrough_text,
                              gsize                text_len,  
                              ConfParser*          cp,
                              GError**             error)
{
	conf_parser_passthrough_func   passthrough = cp->parser->passthrough;

	if (passthrough)
		passthrough (cp, passthrough_text, text_len, cp->data, error);
}

void conf_parser_error (GMarkupParseContext* context,
                        GError*              error,
                        ConfParser*          cp)
{
	conf_parser_error_func   proc_error = cp->parser->error;

	if (proc_error)
		proc_error (cp, error, cp->data);
}

ConfParser* conf_parser_new ()
{
	ConfParser* cp = g_malloc (sizeof (ConfParser));

	conf_parser_init (cp);

	return cp;
}

void conf_parser_destroy (ConfParser* cp)
{
	conf_parser_finalize (cp);
	g_free (cp);
}

void conf_parser_init (ConfParser* cp)
{
	cp->markup_parser.start_element = (markup_parser_start_element_func)conf_parser_start_element;
	cp->markup_parser.end_element = (markup_parser_end_element_func)conf_parser_end_element;
	cp->markup_parser.text  = (markup_parser_text_func)conf_parser_text;
	cp->markup_parser.passthrough = (markup_parser_passthrough_func)conf_parser_passthrough;
	cp->markup_parser.error = (markup_parser_error_func)conf_parser_error;

	cp->parse_context = g_markup_parse_context_new (&cp->markup_parser, 0, cp, NULL);

	cp->parser_stack = g_array_sized_new (FALSE, FALSE, sizeof (ConfParserEntry), 8);
	cp->data_stack = g_ptr_array_sized_new (8);

	cp->parser = &dummy_parser;
	cp->data = NULL;
}

void conf_parser_finalize (ConfParser* cp)
{
	g_markup_parse_context_free (cp->parse_context);
	g_array_free (cp->parser_stack, TRUE);
	g_ptr_array_free (cp->data_stack, TRUE);
}

void conf_parser_push (ConfParser* cp, ConfParserEntry* parser, void* data)
{
	conf_parser_push_parser (cp, parser);
	conf_parser_push_data (cp, data);
}

void conf_parser_pop (ConfParser* cp)
{
	conf_parser_pop_parser (cp);
	conf_parser_pop_data (cp);
}

void conf_parser_push_parser (ConfParser*       cp,
                              ConfParserEntry*  parser)
{
	cp->parser = parser;
	g_array_append_val (cp->parser_stack, *parser);
}

void conf_parser_pop_parser (ConfParser*  cp)
{
	guint len;

	len = cp->data_stack->len;
	if (len) {
		g_array_set_size (cp->parser_stack, --len);
		cp->parser = (len) ? &g_array_index (cp->parser_stack, ConfParserEntry, --len) : &dummy_parser;
	}
}

void conf_parser_push_data (ConfParser*  cp,
                            gpointer     data)
{
	cp->data = data;
	g_ptr_array_add (cp->data_stack, data);
}

void conf_parser_pop_data (ConfParser*  cp)
{
	guint len;

	len = cp->data_stack->len;
	if (len) {
		g_ptr_array_set_size (cp->data_stack, --len);
		cp->data = (len) ? cp->data_stack->pdata[--len] : NULL;
	}
}

gboolean conf_parser_read_file (ConfParser* cp,
                                const gchar* filename,
                                GError** error)
{
	FILE*    file;
	guint8*  buffer;
	guint    size;
	gboolean parse_ok = TRUE;

	file = fopen (filename, "r");
	if (file==NULL)
		return FALSE;

	buffer = g_malloc (4096);

	while ( (size=fread (buffer, 1, 4096, file)) && parse_ok) {
		parse_ok = g_markup_parse_context_parse (cp->parse_context,
		                                         buffer, size, error);
	}

	g_markup_parse_context_end_parse (cp->parse_context, NULL);
	g_free (buffer);
	fclose (file);

	return parse_ok;
}

