#include <glib.h>
#include <curl/curl.h>
#include "urlgfe.h"
#include "download_backend.h"
#include "utility.h"

typedef struct _DownloadBackendCurl DownloadBackendCurl;

struct _DownloadBackendCurl
{
	CURL*   curl;
	gdouble last_completed;
	glong   resume_from;

	DownloadThread* thread;

	char    error_string[CURL_ERROR_SIZE + 1];
};


static int curl_progress_func (DownloadBackendCurl* backend,
                               double dl_total,
                               double dl_now,
                               double ul_total,
                               double ul_now);

char* url_post_encode (const char* pre_url, int len)
{
	GString* gstring;
	UrlPart  urlp;

	if (len==0)
		len = strlen (pre_url);

	url_part (&urlp, pre_url);
	if (urlp.path_beg) {
		gstring = g_string_sized_new (256);
		g_string_append_len (gstring, pre_url, urlp.path_beg - pre_url);
		pre_url = urlp.path_beg;
		for (; len; len--, pre_url++) {
			if ( *pre_url == '\n' )
				continue;
			else if ( *pre_url & 0x80 || *pre_url== ' ')
				g_string_append_printf (gstring, "%c%X", '%', *pre_url);
			else
				g_string_append_c (gstring, *pre_url);
		}
		return g_string_free (gstring, FALSE);
	}
	else
		return g_strndup (pre_url, len);
}

void download_backend_global_init ()
{
	curl_global_init (CURL_GLOBAL_ALL);
}

void download_backend_global_clean ()
{
	curl_global_cleanup ();
}

gpointer download_backend_new (gpointer data)
{
	DownloadBackendCurl* backend;

	backend = g_malloc (sizeof (DownloadBackendCurl));
	backend->curl = curl_easy_init ();
	backend->thread = data;

	return backend;
}

void download_backend_destroy (gpointer data)
{
	DownloadBackendCurl* backend = data;

	curl_easy_cleanup (backend->curl);

	g_free (backend);
}

gint download_backend_run (gpointer backend, DownloadThread* dt)
{
	DownloadBackendCurl* dbc = backend;
	GString* filename = dt->out_filename;
	FILE*    file;
	DownloadSetting* setting = DOWNLOAD_SETTING (dt->download_data);

	CURL*    curl = dbc->curl;
	CURLcode curl_code;
	struct   curl_slist* slist = NULL;
//	long     curl_proxy_type;

	// string data used by libcurl
	// besure curl session finish when you release these.
	gchar*   encoded_url = NULL;
	gchar*   referrer = NULL;
	gchar*   proxy_user_pwd = NULL;

	dbc->last_completed = 0.0;

	// check file
	file = fopen (filename->str, (dt->resumable)? "ab" : "wb");
	if (file==NULL) {
		download_thread_job_set_message (dt, _("File open failure."));
		return DOWNLOAD_BACKEND_ERROR;
	}

	if (dt->resumable) {
		fseek (file, 0, SEEK_END);
		dbc->resume_from = ftell (file);
	} else {
		dbc->resume_from = 0;
	}

	// test code
//	curl_easy_setopt (curl, CURLOPT_UNRESTRICTED_AUTH, 1);
	curl_easy_setopt (curl, CURLOPT_USERAGENT, "Mozilla/4.0 (compatible; MSIE 5.00; Windows 98)");
//	curl_easy_setopt (curl, CURLOPT_USERAGENT, "Internet Explorer");
	curl_easy_setopt (curl, CURLOPT_COOKIEFILE, "");

	// set proxy ----------------------------------------------------
	// clear proxy setting
	curl_easy_setopt (curl, CURLOPT_PROXYTYPE, CURLPROXY_HTTP);
	curl_easy_setopt (curl, CURLOPT_PROXY, NULL);
	curl_easy_setopt (curl, CURLOPT_PROXYUSERPWD, NULL);

	if (PROXY_SETTING (setting)->type != PROXY_TYPE_NONE) {
		if (PROXY_SETTING (setting)->type == PROXY_TYPE_HTTP)
			curl_easy_setopt (curl, CURLOPT_PROXYTYPE, CURLPROXY_HTTP);
		else
			curl_easy_setopt (curl, CURLOPT_PROXYTYPE, CURLPROXY_SOCKS5);
		curl_easy_setopt (curl, CURLOPT_PROXY,
		                  PROXY_SETTING (setting)->address);
		curl_easy_setopt (curl, CURLOPT_PROXYPORT,
		                  (long)PROXY_SETTING (setting)->port);

		if (PROXY_SETTING (setting)->authentication) {
			// set username and password
			proxy_user_pwd = g_strdup_printf ("%s:%s",
			                       STR_N2E (PROXY_SETTING (setting)->username),
			                       STR_N2E (PROXY_SETTING (setting)->password));
			curl_easy_setopt (curl, CURLOPT_PROXYUSERPWD, proxy_user_pwd);
		}
		else if (PROXY_SETTING (setting)->type == PROXY_TYPE_SOCKS5) {
			// Avoid crash when using socks5 without authentication.
			curl_easy_setopt (curl, CURLOPT_PROXYUSERPWD, ":");
		}
	}

	// set redirection
	if (setting->n_redirect) {
		curl_easy_setopt (curl, CURLOPT_FOLLOWLOCATION, 1);
		curl_easy_setopt (curl, CURLOPT_MAXREDIRS, setting->n_redirect);
	} else {
		curl_easy_setopt (curl, CURLOPT_FOLLOWLOCATION, 0);
	}

	// set referrer
//	referrer = g_strdup_printf ("Referrer: %s", STR_N2E (setting->referer));
//	slist = curl_slist_append (slist, referrer);
//	curl_easy_setopt (curl, CURLOPT_HTTPHEADER, slist);
	curl_easy_setopt (curl, CURLOPT_REFERER, STR_N2E (setting->referer));

	// set url
	encoded_url = url_post_encode (STR_N2E (setting->url), 0);
	curl_easy_setopt (curl, CURLOPT_URL, encoded_url);

	// set download data
	curl_easy_setopt (curl, CURLOPT_FILE, file);
	curl_easy_setopt (curl, CURLOPT_NOPROGRESS, FALSE);
	curl_easy_setopt (curl, CURLOPT_PROGRESSFUNCTION, curl_progress_func);
	curl_easy_setopt (curl, CURLOPT_PROGRESSDATA, dbc);
	curl_easy_setopt (curl, CURLOPT_RESUME_FROM, dbc->resume_from);
	curl_easy_setopt (curl, CURLOPT_ERRORBUFFER, dbc->error_string);

	// perform curl
	curl_code = curl_easy_perform (curl);
	fclose (file);
	curl_slist_free_all (slist);

	// free string that used by libcurl
	g_free (referrer);
	g_free (proxy_user_pwd);
	g_free (encoded_url);

//	dt->time_state.
//	set_job_speed (NULL);

	if (curl_code == CURLE_OK) {
		Download* dld = dt->download_data;

		if (dld->completed + dbc->resume_from == dld->total + dbc->resume_from &&
		    dld->completed + dbc->resume_from != 0 )
		{
			return DOWNLOAD_BACKEND_OK;
		}
		else {
			unlink (filename->str);
			download_thread_job_set_message (dt, _("File length error. Try to increase redirection limit."));
			return DOWNLOAD_BACKEND_ERROR;
		}
	}

	printf ("Curl error %.2u: %s\n", curl_code, dbc->error_string);
	download_thread_job_set_message (dt, dbc->error_string);

	// check curl return code
	switch (curl_code) {
		// can resume
		case CURLE_PARTIAL_FILE:
		case CURLE_GOT_NOTHING:
			dt->resumable = TRUE;
			break;
		// can't resume
		case CURLE_HTTP_RANGE_ERROR:
			dt->resumable = FALSE;
			dbc->resume_from = 0;
			break;
		case CURLE_ABORTED_BY_CALLBACK:
			return DOWNLOAD_BACKEND_ABORT;
//		case CURLE_UNSUPPORTED_PROTOCOL:
//		case CURLE_FAILED_INIT:
//		case CURLE_URL_MALFORMAT:
//		case CURLE_URL_MALFORMAT_USER:
//		case CURLE_COULDNT_RESOLVE_PROXY:
//		case CURLE_COULDNT_RESOLVE_HOST:
//		case CURLE_COULDNT_CONNECT:
//		case CURLE_FTP_WEIRD_SERVER_REPLY:
//		case CURLE_FTP_ACCESS_DENIED:
//		case CURLE_FTP_USER_PASSWORD_INCORRECT:
		default:
			unlink (filename->str);
			return DOWNLOAD_BACKEND_ERROR;
	}

	return DOWNLOAD_BACKEND_CAN_RETRY;
}

static int curl_progress_func (DownloadBackendCurl* dbc,
                               double dl_total,
                               double dl_now,
                               double ul_total,
                               double ul_now)
{
	DownloadThread* dt = dbc->thread;
	Download* download_data = dt->download_data;
	gdouble timer = download_thread_get_timer (dt);
	gdouble speed;

	if (dl_now) {
		download_thread_job_set_info_valid (dt, TRUE);
		speed = (dl_now - dbc->last_completed) / timer;
		download_data->speed = (download_data->speed + speed) /2;
		download_data->completed = dl_now + dbc->resume_from;
		download_data->total     = dl_total + dbc->resume_from;
		dbc->last_completed = dl_now;
		// curl pass dl_total=0 sometimes.
		if (download_data->total < download_data->completed)
			download_data->total = download_data->completed;
	}

	// avoid div zero
	if (download_data->total)
		download_data->percent = (download_data->completed / download_data->total)*100;

	download_thread_job_redraw (dt);
	// return true when abort.
	return (dt->thread_state >= DOWNLOAD_THREAD_SKIP);
}


