#include <stdio.h>
#include "utility.h"
#include "urlgfe.h"
#include "download_thread.h"
#include "download_backend.h"

static gpointer download_thread_func (DownloadThread* dt);

#ifdef G_OS_WIN32

#include <windows.h>
void download_thread_retry_delay (DownloadThread* dt)
{
	int count = DOWNLOAD_SETTING (dt->download_data)->retry_delay*2;

	while (dt->thread_state < DOWNLOAD_THREAD_SKIP && count) {
		Sleep (500);  // check state every 0.5 second.
		count--;
	}
}

#else
void download_thread_retry_delay (DownloadThread* dt)
{
	int count = DOWNLOAD_SETTING (dt->download_data)->retry_delay;

	while (dt->thread_state < DOWNLOAD_THREAD_SKIP && count) {
		sleep (1);    // check state every 1 second.
		count--;
	}
}

#endif


DownloadThread* download_thread_new (Category* cate,
                                     GtkTreeIter* iter_download)
{
	DownloadThread* dt = g_malloc (sizeof(DownloadThread));

	g_static_rec_mutex_init (&dt->mutex);

	// list
	dt->prev = NULL;
	dt->next = NULL;

	// category
	dt->category = cate;
	category_ref (cate);

	dt->out_filename = g_string_sized_new (128);

	// DownloadStore
	dt->iter_download = *iter_download;
	download_store_get_data (&dt->category->waiting_store,
	                         &dt->iter_download,
	                         &dt->download_data);
	download_store_set_thread (&dt->category->waiting_store,
	                           &dt->iter_download, dt);
	download_ref (dt->download_data);


	// timer
	dt->timer       = g_timer_new ();
	dt->timer_count = 0.0;
	g_timer_stop (dt->timer);

	dt->thread_state = DOWNLOAD_THREAD_READY;

	category_thread_add (cate, dt);
	return dt;
}

void download_thread_free_resource (DownloadThread* dt)
{
	printf ("thread free : %p\n", dt);

	gdk_threads_enter ();
	category_thread_remove (dt->category, dt);
	category_unref (dt->category);
	gdk_threads_leave ();

	g_string_free (dt->out_filename, TRUE);
	g_timer_destroy (dt->timer);
	g_static_rec_mutex_free (&dt->mutex);

	g_free (dt);
}

void download_thread_activate (DownloadThread* dt)
{
	download_thread_lock(dt);

	if (dt->thread_state == DOWNLOAD_THREAD_READY ) {
		dt->thread_state = DOWNLOAD_THREAD_ACTIVE;

		dt->gthread = g_thread_create ((GThreadFunc)download_thread_func,
		                               dt,TRUE,NULL);
	}

	download_thread_unlock(dt);
}

void download_thread_skip (DownloadThread* dt)
{
	download_thread_lock(dt);

	if (dt->thread_state < DOWNLOAD_THREAD_SKIP )
		dt->thread_state = DOWNLOAD_THREAD_SKIP;

	download_thread_unlock(dt);
}

void download_thread_stop (DownloadThread* dt)
{
	download_thread_lock(dt);

	if (dt->thread_state == DOWNLOAD_THREAD_JOB_ERASED )
		dt->thread_state = DOWNLOAD_THREAD_DESTROY;
	else if (dt->thread_state < DOWNLOAD_THREAD_STOP )
		dt->thread_state = DOWNLOAD_THREAD_STOP;

	download_thread_unlock(dt);
}

void download_thread_job_erased (DownloadThread* dt)
{
	download_thread_lock(dt);

	if (dt->thread_state == DOWNLOAD_THREAD_STOP )
		dt->thread_state = DOWNLOAD_THREAD_DESTROY;
	else if (dt->thread_state < DOWNLOAD_THREAD_JOB_ERASED )
		dt->thread_state = DOWNLOAD_THREAD_JOB_ERASED;

	download_thread_unlock(dt);
}

void download_thread_destroy (DownloadThread* dt)
{
	download_thread_lock(dt);

	if (dt->thread_state < DOWNLOAD_THREAD_DESTROY )
		dt->thread_state = DOWNLOAD_THREAD_DESTROY;

	download_thread_unlock(dt);
}

void download_thread_job_redraw (DownloadThread* dt)
{
	download_store_lock (&dt->category->waiting_store);
	download_thread_lock (dt);

	if (dt->thread_state < DOWNLOAD_THREAD_JOB_ERASED) {
		gdk_threads_enter ();
#ifdef _WIN32
		// avoid crash in GTK+ for Win32 ...
		gtk_widget_queue_draw (GTK_WIDGET (dt->category->waiting_view.self));
#else
		download_store_row_changed (&dt->category->waiting_store,
		                            &dt->iter_download);
#endif
		gdk_threads_leave ();
	}

	download_thread_unlock (dt);
	download_store_unlock (&dt->category->waiting_store);
}

gboolean download_thread_test_file (DownloadThread* dt)
{
	gint     temp;
	gint     directory_len;
	gboolean return_val = FALSE;
	GString* name = dt->out_filename;
	DownloadSetting* setting = DOWNLOAD_SETTING (dt->download_data);
	gchar*   utf8_file_name;
	gchar*   os_dir_name;
	gchar*   os_file_name;

	if (setting->directory==NULL || setting->filename==NULL)
		return FALSE;

	// remove '\n' from url
	temp = strlen (setting->url);
	if (setting->url[temp-1]=='\n')
		setting->url[temp-1] = 0;
	// remove '\n' from directory and filename before convert
	temp = strlen (setting->directory);
	if (setting->directory[temp-1]=='\n')
		setting->directory[temp-1] = 0;
	temp = strlen (setting->filename);
	if (setting->filename[temp-1]=='\n')
		setting->filename[temp-1] = 0;

	os_dir_name = g_filename_from_utf8 (setting->directory, -1,
	                                    NULL, NULL, NULL);
	os_file_name = g_filename_from_utf8 (setting->filename, -1,
	                                     NULL, NULL, NULL);
	g_string_truncate (name, 0);
	g_string_append (name, os_dir_name);

	download_thread_lock (dt);
	if (dt->thread_state < DOWNLOAD_THREAD_SKIP && os_dir_name) {
		if ( name->str[name->len-1] != G_DIR_SEPARATOR )
			g_string_append_c (name, G_DIR_SEPARATOR);
		directory_len = name->len;

		g_string_append (name, os_file_name);

		if (make_path_exist (os_dir_name) && os_file_name) {
			temp = 0;
			do {
				if (g_file_test (name->str, G_FILE_TEST_EXISTS)==FALSE) {
					return_val = TRUE;
					break;
				}

				g_string_truncate (name, directory_len);
				g_string_append_printf (name, "%s.%u",
				                        setting->filename,
				                        temp++);
			} while (temp<100);

			utf8_file_name = g_filename_to_utf8 (name->str + directory_len,
			                                     -1, NULL, NULL, NULL);
			download_setting_set_filename (setting, utf8_file_name);
			g_free (utf8_file_name);
			g_string_append (name, ".ug_");
		}
	}
	download_thread_unlock (dt);

	g_free (os_dir_name);
	g_free (os_file_name);

	download_thread_job_redraw (dt);

	return return_val;
}

gboolean download_thread_load_next_job (DownloadThread* dt)
{
	gboolean return_val = FALSE;

	download_store_lock (&dt->category->waiting_store);
	download_thread_lock (dt);

	if (dt->thread_state==DOWNLOAD_THREAD_JOB_ERASED
	        || dt->thread_state < DOWNLOAD_THREAD_STOP)
	{
		dt->thread_state = DOWNLOAD_THREAD_ACTIVE;
		gdk_threads_enter ();
		if (category_thread_count_request (dt->category) >=0
		    && download_store_find_waiting (&dt->category->waiting_store,
		                                    &dt->iter_download) )
		{
			download_store_get_data (&dt->category->waiting_store,
			                         &dt->iter_download,
			                         &dt->download_data);
			download_store_set_thread (&dt->category->waiting_store,
			                           &dt->iter_download,
			                           dt);
			download_ref (dt->download_data);

			return_val = TRUE;
		}
		gdk_threads_leave ();
	}

	download_thread_unlock (dt);
	download_store_unlock (&dt->category->waiting_store);

	return return_val;
}


gdouble download_thread_get_timer (DownloadThread* dt)
{
	if (dt->timer_count==0) {
		g_timer_start (dt->timer);
		dt->timer_count = 1.0;
	}
	else {
		dt->timer_count = g_timer_elapsed (dt->timer, NULL);
		g_timer_start (dt->timer);
	}

	return dt->timer_count;
}

// ===========================================================================

static gpointer download_thread_func (DownloadThread* dt)
{
	gint      backend_code;
	guint     n_retry;
	gpointer  backend;
	gchar*    temp_string;

	backend = download_backend_new (dt);

	do {
		if (download_thread_test_file (dt)==FALSE) {
			download_thread_job_set_state (dt, DOWNLOAD_STATE_ERROR);
			download_thread_job_set_message (dt, _("Filename error or repeat."));
			download_thread_job_redraw (dt);

			download_store_lock (&dt->category->waiting_store);
			download_thread_lock (dt);
			gdk_threads_enter ();
			download_store_set_thread (&dt->category->waiting_store,
			                           &dt->iter_download, NULL);
			gdk_threads_leave ();
			download_thread_unlock (dt);
			download_store_unlock (&dt->category->waiting_store);

			continue;
		}

		dt->resumable = TRUE;
		n_retry = 0;

		// download & retry loop
		while (1) {
			if (dt->thread_state >= DOWNLOAD_THREAD_SKIP) {
				download_thread_job_set_state (dt, DOWNLOAD_STATE_PAUSE);
				break;
			}
			if (n_retry > DOWNLOAD_SETTING(dt->download_data)->n_retries) {
				download_thread_job_set_state (dt, DOWNLOAD_STATE_ERROR);
				download_thread_job_set_message (dt, _("Retry too many times."));
				break;
			}
			download_thread_job_set_speed (dt, 0);
			download_thread_job_set_state (dt, DOWNLOAD_STATE_EXECUTING);
			download_thread_job_set_retry (dt, n_retry);
			download_thread_job_set_total (dt, 0);
			download_thread_job_set_completed (dt, 0);
			download_thread_job_set_percent (dt, 0);
			download_thread_job_redraw (dt);
			n_retry++;

			dt->timer_count = 0.0;
			// begin download here
			backend_code = download_backend_run (backend, dt);

			// check download backend return code
			if (backend_code==DOWNLOAD_BACKEND_ABORT) {
				download_thread_job_set_state (dt, DOWNLOAD_STATE_PAUSE);
				break;
			}
			else if (backend_code==DOWNLOAD_BACKEND_OK) {
				temp_string = g_strndup (dt->out_filename->str,
				                         dt->out_filename->len-4);
				rename (dt->out_filename->str, temp_string);
				g_free (temp_string);
				download_thread_job_set_state (dt, DOWNLOAD_STATE_COMPLETED);
				break;
			}
			else if (backend_code==DOWNLOAD_BACKEND_ERROR) {
				download_thread_job_set_state (dt, DOWNLOAD_STATE_ERROR);
				break;
			}

			download_thread_job_set_state (dt, DOWNLOAD_STATE_RETRY);
			download_thread_job_redraw (dt);
			download_thread_retry_delay (dt);
		}
		download_thread_job_redraw (dt);

		// do some thing after download
		download_unref (dt->download_data);

		download_store_lock (&dt->category->waiting_store);
		download_thread_lock (dt);
		if (dt->thread_state < DOWNLOAD_THREAD_JOB_ERASED)
		{
			gdk_threads_enter ();
			download_store_set_thread (&dt->category->waiting_store,
			                           &dt->iter_download, NULL);
			if (backend_code==DOWNLOAD_BACKEND_OK) {
				download_store_move_to_store (&dt->category->waiting_store,
				                              &dt->category->completed_store,
				                              &dt->iter_download);
				category_changed_with_children (dt->category, CATEGORY_COMPLETED);
			}
			gdk_threads_leave ();
		}
		download_thread_unlock (dt);
		download_store_unlock (&dt->category->waiting_store);

	} while (download_thread_load_next_job (dt));

	download_backend_destroy (backend);
	download_thread_free_resource (dt);
	return NULL;
}

